import * as React from 'react'
import { Box, Button, Popper, TableRow } from '@mui/material'
import ConfirmDialog from 'components/confirmDialog'
import { Identity } from 'interfaces'
import { useEffect, useState } from 'react'
import { fetchJsonOpts, submitJsonOpts } from 'utils'
import CheckIcon from '@mui/icons-material/Check'
import { Trans, useTranslation } from 'react-i18next'
import { TableCell } from 'components/table'
import { appInst, disableNinVerification, enableIgaCheck } from 'appConfig'

interface IdentityLineProps {
  text: string
  identity: Identity | null
  reloadGuest: () => void
  reloadGuests: () => void
}

const IdentityLine = ({
  text,
  identity,
  reloadGuest,
  reloadGuests,
}: IdentityLineProps) => {
  // Make a line with a confirmation button if the identity has not been verified
  if (identity == null) {
    return <></>
  }
  const [confirmOpen, setConfirmOpen] = useState(false)
  const [t] = useTranslation('common')
  const verifyIdentity = (id: string) => async () => {
    await fetch(`/api/ui/v1/identity/${id}`, submitJsonOpts('PATCH', {}))
    reloadGuest()
    reloadGuests()
  }
  const [verifiedBoxAnchor, setVerifiedBoxAnchor] =
    React.useState<null | HTMLElement>(null)
  const verifiedBoxOpen = Boolean(verifiedBoxAnchor)
  const verifiedBoxId = verifiedBoxOpen ? 'verified-popper' : undefined

  const handleVerifiedInfo = (event: React.MouseEvent<HTMLElement>) => {
    setVerifiedBoxAnchor(verifiedBoxAnchor ? null : event.currentTarget)
  }

  const [identityCheckText, setIdentityCheckText] = useState<string | null>(
    null
  )
  const checkIdentity = (id: string) => {
    const igaMatch = fetch(`/api/ui/v1/identitycheck/${id}`, fetchJsonOpts())
      .then((res) => res.json())
      .then((data) =>
        data.match ? `${data.match.first} ${data.match.last}` : null
      )
      .catch(() => 'failure')
    return igaMatch
  }

  const fetchIdentityCheckText = async (id: string) => {
    checkIdentity(id).then((idtxt) => setIdentityCheckText(idtxt))
  }

  const IdentityCheckText = () => {
    if (identityCheckText === 'failure') {
      return (
        <>
          <Trans t={t} i18nKey="guestInfo.identityCheck.failure" />
          <br />
          <br />
        </>
      )
    }
    if (identityCheckText) {
      return (
        <>
          <Trans t={t} i18nKey="guestInfo.identityCheck.text">
            Warning: {identityCheckText} already exist in IGA with this ID.
          </Trans>
          <br />
          <br />
        </>
      )
    }
    return <></>
  }
  const getDialogText = () => {
    switch (appInst) {
      case 'uio':
        return t('confirmationDialog.text.uio')
      case 'uib':
        return t('confirmationDialog.text.uib')
      default:
        return t('confirmationDialog.text.default')
    }
  }

  useEffect(() => {
    if (enableIgaCheck && confirmOpen && identityCheckText === null) {
      fetchIdentityCheckText(identity.id)
    }
  }, [confirmOpen])
  return (
    <TableRow>
      <TableCell align="left">{text}</TableCell>
      <TableCell align="left">
        <Box
          sx={{
            display: 'flex',
            flexDirection: { xs: 'column', md: 'row' },
            justifyContent: 'flex-start',
          }}
        >
          <Box>
            {identity ? identity.value : ''}
            {!identity.verified ? (
              <>
                <Button
                  aria-label={t('button.verify')}
                  sx={{
                    alignSelf: { xs: 'auto', md: 'flex-end' },
                    marginLeft: { xs: '0rem', md: '1rem' },
                    marginTop: { xs: '0.3rem', md: '0rem' },
                  }}
                  onClick={() => setConfirmOpen(true)}
                  disabled={
                    !identity ||
                    (disableNinVerification &&
                      identity.type === 'norwegian_national_id_number')
                  }
                >
                  {t('button.verify')}
                </Button>
                <ConfirmDialog
                  title={t('confirmationDialog.confirmIdentityTitle')}
                  open={confirmOpen}
                  setOpen={setConfirmOpen}
                  onConfirm={verifyIdentity(identity.id)}
                >
                  <IdentityCheckText />
                  {getDialogText()}
                </ConfirmDialog>
              </>
            ) : (
              <>
                <CheckIcon
                  sx={{
                    fill: (theme) => theme.palette.success.main,
                    verticalAlign: 'text-bottom',
                    border: 0,
                  }}
                  titleAccess={t('guestInfo.verified')}
                />
              </>
            )}
            <Button
              onClick={handleVerifiedInfo}
              sx={{
                alignSelf: { xs: 'auto', md: 'flex-end' },
                marginLeft: { xs: '0rem', md: '1rem' },
                marginTop: { xs: '0.3rem', md: '0rem' },
              }}
            >
              {t('guestInfo.viewInfo')}
            </Button>
            <Popper
              id={verifiedBoxId}
              open={Boolean(verifiedBoxAnchor)}
              anchorEl={verifiedBoxAnchor}
            >
              <Box sx={{ border: 1, p: 1, bgcolor: 'background.paper' }}>
                {identity ? `${t('guestInfo.source')}: ${identity.source}` : ''}
                <br />
                {identity && identity.verified !== ''
                  ? `${t('guestInfo.verifier')}: ${
                      identity.verified === 'manual'
                        ? identity.verified_by
                        : identity.source
                    }`
                  : ''}
              </Box>
            </Popper>
          </Box>
        </Box>
      </TableCell>
    </TableRow>
  )
}

export default IdentityLine
