"""
Collection of tasks designed to be run by django-q
"""

import datetime
from collections import defaultdict

from django.conf import settings

from greg.importers.orgreg import OrgregImporter
from greg.models import Role, Sponsor
from gregui.mailutils.role_ending import RolesEnding


def notify_sponsors_roles_ending() -> list[str]:
    """
    This task notifies sponsors of roles (that are accessible to them)
    that are about to expire.

    It is meant to be run at set intervals. Typically every week or
    two. The script will notify of all roles ending between today and
    the number of days set in the NOTIFIER_LIMIT setting.

    While it is sensible to set the run interval to the same value as
    the NOTIFIER_LIMIT, it is not mandatory, as there is a use case for
    setting the run interval to e.g 1 week and the limit to 2. This
    would ensure that the sponsor gets 2 notifications about the same
    role ending. (Be careful not to swap them though, as that would
    ensure some roles ending without a notification being sent to the
    sponsor).
    """
    # Map ending roles to units
    today = datetime.date.today()
    ending_roles = Role.objects.filter(
        end_date__gte=today,
        end_date__lte=today + datetime.timedelta(days=settings.NOTIFIER_LIMIT),
    )
    unit2roles = defaultdict(list)
    for role in ending_roles:
        unit2roles[role.orgunit.id].append(role)

    # Map sponsors with email to units
    sponsors = Sponsor.objects.filter(work_email__isnull=False)
    sp2unit = {s.id: list(s.get_allowed_units()) for s in sponsors}

    # Map sponsors to ending roles
    sp2roles = defaultdict(list)
    for sp, units in sp2unit.items():
        for unit in units:
            for role in unit2roles[unit.id]:
                sp2roles[sp].append(role)

    # Send emails to sponsors
    remindermailer = RolesEnding()
    task_ids = []
    for sp, roles in sp2roles.items():
        task_ids.append(
            remindermailer.queue_mail(
                mail_to=sponsors.get(id=sp).work_email,  # type: ignore
                num_roles=len(roles),
            )
        )
    return task_ids


def import_from_orgreg():
    importer = OrgregImporter()
    importer.handle()
