import datetime
import pytest
from django.conf import settings
from django.core.management import call_command

from greg.models import Sponsor, SponsorOrganizationalUnit


class MockResponse:
    def __init__(self, response):
        self.response = response

    def json(self):
        return self.response


@pytest.fixture
def group_response():
    created_at = (datetime.datetime.now() - datetime.timedelta(days=1)).isoformat()
    return MockResponse(
        f"""{{
            "href": "",
            "id": 1,
            "name": "",
            "description": "",
            "created_at": "{created_at}",
            "contexts": ["das"],
            "visibility": "",
            "moderators": ""
        }}"""
    )


@pytest.fixture
def group_members_response():
    return MockResponse("""{"members":[{"href": "asd", "type": "person", "id": 1}]}""")


@pytest.fixture
def group_members_response2():
    return MockResponse("""{"members":[{"href": "asd", "type": "person", "id": 2}]}""")


@pytest.fixture
def person_response():
    return MockResponse(
        """{
        "contexts": [""],
        "href": "",
        "names": [
            {"source_system":"DFO_SAP", "variant": "FIRST", "name": "Ola"},
            {"source_system":"DFO_SAP", "variant": "LAST", "name": "Nordmann"}
        ],
        "id": 1}"""
    )


@pytest.fixture
def person_response2():
    return MockResponse(
        """{
        "contexts": [""],
        "href": "",
        "names": [
            {"source_system":"DFO_SAP", "variant": "FIRST", "name": "Kari"},
            {"source_system":"DFO_SAP", "variant": "LAST", "name": "Nordmann"}
        ],
        "id": 2}"""
    )


@pytest.fixture
def account_response():
    return MockResponse(
        """{"accounts": [{
        "href": "",
        "primary": true,
        "id": 1,
        "name": "olanord"}]}"""
    )


@pytest.fixture
def account_response2():
    return MockResponse(
        """{"accounts": [{
        "href": "",
        "primary": true,
        "id": 2,
        "name": "karinor"}]}"""
    )


@pytest.mark.django_db
def test_import_sponsors_from_cerebrum(
    requests_mock,
    group_response,
    group_members_response,
    group_members_response2,
    person_response,
    person_response2,
    account_response,
    account_response2,
    sponsor_org_unit,
    unit_foo2,
):
    """Check that only non manual sponsors are imported and linked"""
    settings.CEREBRUM_CLIENT = {
        "url": "http://example.com/cerebrum/",
        "headers": {"X-Gravitee-Api-Key": "fake-key"},
    }
    settings.CEREBRUM_MANUAL_SPONSOR_UNITS = ["12345"]
    requests_mock.get(
        "http://example.com/cerebrum/v1/groups/adm-leder-12345",
        text=group_response.json(),
    )
    requests_mock.get(
        "http://example.com/cerebrum/v1/groups/adm-leder-12345/members/",
        text=group_members_response.json(),
    )
    requests_mock.get(
        "http://example.com/cerebrum/v1/persons/1",
        text=person_response.json(),
    )
    requests_mock.get(
        "http://example.com/cerebrum/v1/persons/1/accounts",
        text=account_response.json(),
    )

    requests_mock.get(
        "http://example.com/cerebrum/v1/groups/adm-leder-23456",
        text=group_response.json(),
    )
    requests_mock.get(
        "http://example.com/cerebrum/v1/groups/adm-leder-23456/members/",
        text=group_members_response2.json(),
    )
    requests_mock.get(
        "http://example.com/cerebrum/v1/persons/2",
        text=person_response2.json(),
    )
    requests_mock.get(
        "http://example.com/cerebrum/v1/persons/2/accounts",
        text=account_response2.json(),
    )
    assert Sponsor.objects.all().count() == 1
    assert SponsorOrganizationalUnit.objects.count() == 1
    call_command("import_sponsors_from_cerebrum")
    # One from unit with one sponsor
    # One excluded from unit with stedkode in manual list
    assert Sponsor.objects.all().count() == 2
    assert SponsorOrganizationalUnit.objects.count() == 2
