from smtplib import SMTPRecipientsRefused
from django.conf import settings
from django.core import mail
from django.template.context import Context

import pytest
from gregui.mailutils.confirm_guest import ConfirmGuest

from gregui.mailutils.invite_guest import InviteGuest, try_send_registration_mail


@pytest.mark.django_db
def test_registration_template(registration_template):
    prefilled_template = {
        "from_email": None,
        "recipient_list": ["test@example.com"],
        "subject": "registration subject",
        "message": """Dette er en automatisk generert melding fra gjesteregistreringstjenesten.
Du har blitt registrert som gjest på local av Foo Bar.
For å fullføre registreringen av gjestekontoen følg denne lenken: Baz

This message has been automatically generated by the guest registration system.
You have been registered as a guest at local by Foo Bar.
To complete the registration of your guest account, please follow this link: Baz""",
    }
    ig = InviteGuest()
    template = ig.get_template()
    context = Context(
        {
            "institution": settings.INSTANCE_NAME,
            "sponsor": "Foo Bar",
            "registration_link": "Baz",
        }
    )
    assert template.get_subject(context) == prefilled_template["subject"]
    assert template.get_body(context) == prefilled_template["message"]
    assert template.from_email == prefilled_template["from_email"]


@pytest.mark.django_db
def test_confirmation_template(confirmation_template):
    prefilled_template = {
        "from_email": None,
        "recipient_list": ["test@example.com"],
        "subject": "confirmation subject",
        "message": """Dette er en automatisk generert melding fra gjesteregistreringstjenesten.
Din gjest, Foo Bar, har fullført registrering, bekreft gjesten her: Baz

This message has been automatically generated by the guest registration system.
Your guest, Foo Bar, has completed their registration, please confirm the guest here: Baz""",
    }
    scg = ConfirmGuest()
    template = scg.get_template()
    context = Context(
        {
            "guest": "Foo Bar",
            "confirmation_link": "Baz",
        }
    )
    assert template.get_subject(context) == prefilled_template["subject"]
    assert template.get_body(context) == prefilled_template["message"]
    assert template.from_email == prefilled_template["from_email"]


@pytest.mark.django_db
def test_registration_mail(registration_template):
    mail.outbox = []
    try_send_registration_mail(
        guest_name="Guest Guesterson",
        guest_mail="test@example.no",
        sponsor_id=1,
        sponsor_name="Foo",
        token="secret-key",
    )
    assert len(mail.outbox) == 1
    assert mail.outbox[0].to == ["test@example.no"]


@pytest.mark.django_db
def test_confirmation_mail(confirmation_template):
    mail.outbox = []
    scg = ConfirmGuest()
    task_id = scg.queue_mail(
        mail_to="test@example.no", guest_name="Foo", guest_person_id=123
    )
    assert isinstance(task_id, str)


@pytest.mark.django_db
def test_send_invite_mail(registration_template, invited_person):
    """Verify function queues an email when called"""
    _, link = invited_person
    ig = InviteGuest()
    task_id = ig.queue_mail(link)
    assert isinstance(task_id, str)


@pytest.mark.django_db
def test_send_invite_mail_no_mail(invited_person):
    """Verify function returns None if an email is not present"""
    person, link = invited_person
    person.private_email.delete()
    ig = InviteGuest()
    task_id = ig.queue_mail(link)
    assert task_id is None


@pytest.mark.django_db
def test_confirmation_mail_from_link(invited_person, confirmation_template):
    mail.outbox = []
    _, link = invited_person
    assert len(mail.outbox) == 0
    scg = ConfirmGuest()
    task_id = scg.send_confirmation_mail_from_link(link)
    assert isinstance(task_id, str)


@pytest.mark.django_db
def test_try_send_registration_mail(
    registration_template, invalid_email_template, mocker, sponsor_foo
):
    send_invite_mock_function = mocker.patch("django.core.mail.send_mail")

    def side_effect(*args, **kwargs):
        if "oladunk@example.com" in kwargs["recipient_list"]:
            raise SMTPRecipientsRefused(recipients=kwargs["recipient_list"])

    send_invite_mock_function.side_effect = side_effect

    try_send_registration_mail(
        guest_name="Ola Dunk",
        guest_mail="oladunk@example.com",
        sponsor_id=sponsor_foo.id,
        sponsor_name=sponsor_foo.first_name + " " + sponsor_foo.last_name,
        token="secret-key",
    )
    assert send_invite_mock_function.call_count == 2
    send_invite_mock_function.assert_called_with(
        **{
            "subject": "invalid email",
            "message": "Your guest Ola Dunk has bad email oladunk@example.com",
            "from_email": None,
            "recipient_list": ["foo@example.org"],
        }
    )
